#!/usr/bin/env python
# coding: utf-8

#########################################################################
#########################################################################

"""
   File Name: _cext.pyx
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Wed Nov  5 16:32:11 2014 CST
"""
DESCRIPTION = """
"""

# import both numpy and the Cython declarations for numpy
import numpy as np
cimport numpy as cnp
from libc.stdint cimport int8_t, uint8_t, \
    int16_t, uint16_t, int32_t, uint32_t
import operator
from itertools import izip as zip

from hdidx import _cext as cext

# use the Numpy-C-API from Cython
cnp.import_array()

# cdefine the signature of our c function
cdef extern from "mih.h":
    int get_keys_dist(uint32_t slice, int len,
                      int dist, uint32_t * keys)

cdef extern from "hamdist.h":
    uint16_t hamdist(uint8_t * qry, uint8_t * db, int dim)


def c(n,k):
    return reduce(operator.mul, range(n - k + 1, n + 1)) /\
        reduce(operator.mul, range(1, k +1)) if k > 0 else 1

# create the wrapper code, with numpy type annotations
def get_key_map(nbits):
    if nbits > 32:
        raise StandardError("Number of bits cannot exceed 32!")
    out = []
    for d in xrange(nbits+1):
        codes = np.zeros(c(nbits, d), dtype=np.uint32)
        num = get_keys_dist(0, nbits, d,
                            <uint32_t *> cnp.PyArray_DATA(codes))
        if num != codes.shape[0]:
            raise Exception("Memory leak!!!")
        out.append(codes)
    return out

def get_keys_16bit(codes):
    out = np.zeros((codes.shape[0], codes.shape[1]/2), np.uint16)
    for i in xrange(codes.shape[0]):
        for j in xrange(0, codes.shape[1], 2):
            out[i, j/2] = codes[i, j] * 256 + codes[i, j+1]
    return out

def search(qry_code, db_codes, tables, key_map, topk):
    pass

def search_for_sub_dist(sub_dist, qry_keys, qry_code, proced,
                        db_codes, idmap,
                        tables, key_map, ret_set):
    nbits = qry_code.shape[-1]
    for table, subcode in zip(tables, qry_keys[0]):
        for mask in key_map[sub_dist]:
            for cur_id in table.get(mask ^ subcode, []):
                # if not proced.test(cur_id):
                #     proced.set(cur_id)
                if cur_id not in proced:
                    proced.add(cur_id)
                    dist = hamdist(<uint8_t *> cnp.PyArray_DATA(qry_code),
                                   <uint8_t *> cnp.PyArray_DATA(db_codes[idmap[cur_id]]),
                                   nbits)
                    ret_set[dist].append(cur_id)

cdef extern from "mih.h":
    cdef cppclass MultiIndexer:
        MultiIndexer(int, int, int)
        int get_num_items()
        int add(uint8_t * codes, int num)
        int search(uint8_t * query, int32_t * ids,
                   int16_t * dis, int topk)
        int load(char * codes)
        int save(char * codes)

cdef class PyMultiIndexer:
    cdef MultiIndexer * thisptr

    def __cinit__(self, int nbits, int ntbls, int capacity=0):
        self.thisptr = new MultiIndexer(nbits, ntbls, capacity)

    def __dealloc__(self):
        del self.thisptr

    def get_num_items(self):
        return self.thisptr.get_num_items()

    def add(self, cnp.ndarray[uint8_t, ndim=2, mode="c"] codes):
        return self.thisptr.add(<uint8_t *> cnp.PyArray_DATA(codes),
                                codes.shape[0])

    def search(self, cnp.ndarray[uint8_t, ndim=2, mode="c"] qry, int topk):
        ids = np.zeros(topk, np.int32)
        dis = np.zeros(topk, np.int16)
        self.thisptr.search(<uint8_t *> cnp.PyArray_DATA(qry),
                            <int32_t *> cnp.PyArray_DATA(ids),
                            <int16_t *> cnp.PyArray_DATA(dis),
                            topk)
        return ids, dis

    def load(self, idx_path):
        return self.thisptr.load(<char *> idx_path)

    def save(self, idx_path):
        return self.thisptr.save(<char *> idx_path)



#!/usr/bin/env python
# coding: utf-8

"""
   File Name: hamming.py
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Mon Jul 27 10:22:06 2015 CST
"""
DESCRIPTION = """
Indexers for binary codes in Hamming space.
"""

import os
import logging

import numpy as np

import hdidx.encoder
from hdidx.indexer import Indexer
from hdidx.util import Profiler
from hdidx.storage import createStorage

import hdidx._cext as cext
from hdidx import _mih as mih

BIT_CNT_MAP = np.array([bin(i).count("1") for i in xrange(256)], np.uint16)
DEFAULT_HAMMING_ENCODER = "SHEncoder"


class SHIndexer(Indexer):
    def __init__(self, encoder=DEFAULT_HAMMING_ENCODER):
        Indexer.__init__(self)
        self.encoder = getattr(hdidx.encoder, encoder)()
        self.set_storage()

    def __del__(self):
        pass

    def build(self, pardic=None):
        self.encoder.build(pardic)

    def set_storage(self, storage_type='mem', storage_parm=None):
        self.storage = createStorage(storage_type, storage_parm)

    def add(self, vals, keys=None):
        num_vals = vals.shape[0]
        if keys is None:
            num_base_items = self.storage.get_num_items()
            keys = np.arange(num_base_items, num_base_items + num_vals,
                             dtype=np.int32)
        else:
            keys = np.array(keys, dtype=np.int32).reshape(-1)

        start_id = 0
        for start_id in range(0, num_vals, self.BLKSIZE):
            cur_num = min(self.BLKSIZE, num_vals - start_id)
            logging.info("%8d/%d: %d" % (start_id, num_vals, cur_num))
            codes = self.encoder.encode(vals[start_id:start_id+cur_num, :])
            self.storage.add(codes, keys[start_id:start_id+cur_num])

    def remove(self, keys):
        raise Exception(self.ERR_UNIMPL)

    @staticmethod
    def hammingDist(B1, B2):
        """
        Compute hamming distance between two sets of samples (B1, B2)

        Dh=hammingDist(B1, B2);

        Input
        B1, B2: compact bit vectors. Each datapoint is one row.
        size(B1) = [ndatapoints1, nwords]
        size(B2) = [ndatapoints2, nwords]
        It is faster if ndatapoints1 < ndatapoints2

        Output
        Dh = hamming distance.
        size(Dh) = [ndatapoints1, ndatapoints2]

        example query
        Dhamm = hammingDist(B2, B1);
        this will give the same result than:
        Dhamm = distMat(U2>0, U1>0).^2;
        the size of the distance matrix is:
        size(Dhamm) = [Ntest x Ntraining]
        """

        if B1.ndim == 1:
            B1 = B1.reshape((1, -1))

        if B2.ndim == 1:
            B2 = B2.reshape((1, -1))

        npt1, dim1 = B1.shape
        npt2, dim2 = B2.shape

        if dim1 != dim2:
            raise Exception("Dimension not consists: %d, %d" % (dim1, dim2))

        Dh = np.zeros((npt1, npt2), np.uint16)

        for i in xrange(npt1):
            Dh[i, :] = BIT_CNT_MAP[np.bitwise_xor(B1[i, :], B2)].sum(1)

        return Dh

    @staticmethod
    def hammingDist2(B1, B2):
        """
        Compute hamming distance between two sets of samples (B1, B2)

        Dh=hammingDist(B1, B2);

        Input
        B1, B2: compact bit vectors. Each datapoint is one row.
        size(B1) = [ndatapoints1, nwords]
        size(B2) = [ndatapoints2, nwords]
        It is faster if ndatapoints1 < ndatapoints2

        Output
        Dh = hamming distance.
        size(Dh) = [ndatapoints1, ndatapoints2]

        example query
        Dhamm = hammingDist(B2, B1);
        this will give the same result than:
        Dhamm = distMat(U2>0, U1>0).^2;
        the size of the distance matrix is:
        size(Dhamm) = [Ntest x Ntraining]
        """

        if B1.ndim == 1:
            B1 = B1.reshape((1, -1))

        if B2.ndim == 1:
            B2 = B2.reshape((1, -1))

        npt1, dim1 = B1.shape
        npt2, dim2 = B2.shape

        if dim1 != dim2:
            raise Exception("Dimension not consists: %d, %d" % (dim1, dim2))

        Dh = cext.hamming(B1, B2)

        return Dh

    def search(self, queries, topk=None, **kwargs):
        nq = queries.shape[0]
        nbits = self.encoder.ecdat['nbits']

        # qry_codes = self.encoder.encode(queries)
        db_codes = self.storage.get_codes()
        idsquerybase = self.storage.get_keys()

        dis = np.ones((nq, topk), np.single) * np.inf
        ids = np.ones((nq, topk), np.int32) * -1

        profiler = Profiler()
        interval = 100 if nq >= 100 else 10
        time_total = 0.0    # total time for all queries
        logging.info('Start Querying ...')
        for qry_id in range(nq):
            profiler.start("encoding")  # time for computing the distances
            qry_code = self.encoder.encode(queries[qry_id:qry_id+1])
            profiler.end()

            profiler.start("distance")  # time for computing the distances
            disquerybase = self.hammingDist2(qry_code, db_codes).reshape(-1)
            profiler.end()

            profiler.start("knn")       # time for finding the kNN
            cur_ids = cext.knn_count(disquerybase, nbits, topk)
            profiler.end()

            profiler.start("result")    # time for getting final result
            ids[qry_id, :] = idsquerybase[cur_ids]
            dis[qry_id, :] = disquerybase[cur_ids]
            profiler.end()

            if (qry_id+1) % interval == 0:
                time_total += profiler.sum_overall()
                logging.info(
                    '\t%d/%d: %.3fms per query' %
                    (qry_id+1, nq, profiler.sum_average() * 1000))
                logging.info("\t\t%s" % profiler.str_average())
                profiler.reset()
        logging.info('Querying Finished!')
        time_total += profiler.sum_overall()
        logging.info("Average querying time: %.3fms" % (time_total * 1000 / nq))

        return ids, dis


class MIHIndexer(Indexer):
    def __init__(self, encoder=DEFAULT_HAMMING_ENCODER):
        Indexer.__init__(self)
        self.encoder = getattr(hdidx.encoder, encoder)()
        self.indexer = None
        self.key_map = mih.get_key_map(16)
        self.set_storage()

    def __del__(self):
        pass

    def build(self, pardic=None):
        self.encoder.build(pardic)

    def set_storage(self, storage_type='mem', storage_parm=None):
        self.idx_path = storage_parm['path'] if storage_parm else None

        if self.idx_path is not None:
            nbits = self.encoder.ecdat['nbits']
            self.ntbls = nbits / 16
            self.indexer = mih.PyMultiIndexer(nbits, self.ntbls, 1000000)

            if os.path.exists(self.idx_path):
                self.indexer.load(self.idx_path)

    def add(self, vals, keys=None):
        num_vals = vals.shape[0]
        if keys is None:
            num_base_items = self.indexer.get_num_items()
            keys = np.arange(num_base_items, num_base_items + num_vals,
                             dtype=np.int32)
        else:
            keys = np.array(keys, dtype=np.int32).reshape(-1)

        start_id = 0
        for start_id in range(0, num_vals, self.BLKSIZE):
            cur_num = min(self.BLKSIZE, num_vals - start_id)
            logging.info("%8d/%d: %d" % (start_id, num_vals, cur_num))
            codes = self.encoder.encode(vals[start_id:start_id+cur_num, :])
            self.indexer.add(codes)
        logging.info("%8d/%d (Done!)" % (num_vals, num_vals))

        self.indexer.save(self.idx_path)

    def remove(self, keys):
        raise Exception(self.ERR_UNIMPL)

    def search(self, queries, topk=None, **kwargs):
        nq = queries.shape[0]
        # qry_codes = self.encoder.encode(queries)

        dis = np.ones((nq, topk), np.single) * np.inf
        ids = np.ones((nq, topk), np.int32) * -1

        profiler = Profiler()
        interval = 100 if nq >= 100 else 10
        time_total = 0.0    # total time for all queries
        logging.info('Start Querying ...')
        for qry_id in range(nq):
            profiler.start("encoding")    # time for getting final result
            qry_code = self.encoder.encode(queries[qry_id:qry_id+1])
            profiler.end()

            profiler.start("search")    # time for getting final result
            cur_ids, cur_dis = self.indexer.search(qry_code, topk)
            profiler.end()

            profiler.start("result")    # time for getting final result
            ids[qry_id, :] = cur_ids
            dis[qry_id, :] = cur_dis
            profiler.end()

            if (qry_id+1) % interval == 0:
                time_total += profiler.sum_overall()
                logging.info(
                    '\t%d/%d: %.3fms per query' %
                    (qry_id+1, nq, profiler.sum_average() * 1000))
                logging.info("\t\t%s" % profiler.str_average())
                profiler.reset()
        logging.info('Querying Finished!')
        time_total += profiler.sum_overall()
        logging.info("Average querying time: %.3fms" % (time_total * 1000 / nq))

        return ids, dis
